"""
This module helps us test without having to install the package"""

import os
import sys

addpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, addpath)

import jacobsjsondoc



try:
    from .document import create_document
except ModuleNotFoundError:
    import sys
    if 'setup' not in sys.modules['__main__'].__file__:
        raise
from . import _version

__version__ = _version.__version__
__version_info__ = _version.__version_info__

from .loader import PrepopulatedLoader

def parse(text_data):
    ppl = PrepopulatedLoader()
    ppl.prepopulate(None, text_data)
    doc = create_document(uri=None, loader=ppl)
    return doc



from urllib.parse import urlparse, ParseResult as UrlParseResult
from collections import UserDict
from typing import Union

class JsonAnchor:

    def __init__(self, scheme, netloc, path, fragment):
        self.scheme = scheme
        self.netloc = netloc
        self.path = path
        self.fragment = fragment
    
    @property
    def uri(self):
        if self.scheme and self.netloc:
            return f"{self.scheme}://{self.netloc}{self.path}"
        return self.path
    
    @classmethod
    def from_url_parsed_result(cls, result:UrlParseResult):
        return cls(result.scheme, result.netloc, result.path, result.fragment)

    @classmethod
    def from_string(cls, input:str):
        result = urlparse(input)
        return cls.from_url_parsed_result(result)

    @classmethod
    def empty(cls):
        return cls('', '', '', '')

    def __repr__(self):
        fragment = f"#{self.fragment}" if self.fragment else ""
        if self.path:
            return f"{self.uri}{fragment}"
        return fragment

    def copy(self):
        return self.__class__(self.scheme, self.netloc, self.path, self.fragment)

    def append_to_fragment(self, part):
        self.fragment = f"{self.fragment}/{part}"
        return self

    def change_to(self, result:Union[UrlParseResult,str]):
        new_ref = result
        if isinstance(result, str):
            new_ref = self.from_string(result)
        if new_ref.scheme and new_ref.netloc:
            self.scheme = new_ref.scheme
            self.netloc = new_ref.netloc
        if new_ref.path:
            self.path = new_ref.path
        if new_ref.fragment:
            self.fragment = new_ref.fragment
        return self
    
    def __eq__(self, other):
        alt = other
        if isinstance(other, str):
            alt = self.from_string(other)
        return (self.uri == alt.uri) and (self.fragment == alt.fragment)

    def __hash__(self):
        return self.__repr__().__hash__()


class ReferenceDictionary(UserDict):
    
    def get(self, dollar_id:JsonAnchor):
        return self[dollar_id]

    def put(self, dollar_id:JsonAnchor, node):
        self[dollar_id] = node
        return self



import unittest
import os.path

import jacobsjsondoc

JSON_WITH_A_TAB = """
{
    "things": [
        { "foo": "bar" },
		{ "hello": "world - this line has a tab character instead of spaces" },
        null,
        True,
        False
    ],
    "hi": "💩"
}
"""

class TestParsedTypes(unittest.TestCase):

    def setUp(self):
        self.doc = jacobsjsondoc.parse(JSON_WITH_A_TAB)

    def test_parse_dict(self):
        self.assertIsInstance(self.doc, dict)
        self.assertIsInstance(self.doc["things"][0], dict)
        self.assertIsNone(self.doc["things"][2])

    def test_parse_list(self):
        self.assertIsInstance(self.doc["things"], list)

    def test_parse_booleans(self):
        self.assertTrue(self.doc["things"][3])
        self.assertFalse(self.doc["things"][4])
        self.assertTrue(self.doc["things"][3] is True)
        self.assertTrue(self.doc["things"][4] is False)

    def test_grapheme(self):
        self.assertIsInstance(self.doc["hi"], str)
        self.assertEqual(self.doc["hi"], "💩")

class TestMinLength(unittest.TestCase):

    def setUp(self):
        data_file = os.path.join(os.path.dirname(__file__), "minLength.json")
        data = open(data_file, "r").read()
        self.doc = jacobsjsondoc.parse(data)

    def test_grapheme(self):
        self.assertIsInstance("💩", str)
        self.assertEqual(len("💩"), 1)
        self.assertIsInstance(self.doc["data"], str)
        self.assertIsInstance(self.doc["data"], jacobsjsondoc.document.DocString)
        #self.assertEqual(self.doc[0]["tests"][4]["data"], "💩")
        self.assertEqual(len(self.doc["data"]), 1)





import json
from typing import Optional

from .loader import LoaderBaseClass, FilesystemLoader
from .parser import Parser
from .reference import ReferenceDictionary, JsonAnchor
from .options import ParseOptions, RefResolutionMode


class ReferenceResolutionError(Exception):
    pass

class PathReferenceResolutionError(ReferenceResolutionError):

    def __init__(self, doc, path):
        super().__init__(f"Could not resolve path: '{path}' from {doc.uri}")


class CircularDependencyError(ReferenceResolutionError):
    def __init__(self, uri):
        super().__init__(f"Circular dependency detected when trying to load '{uri}' a second time")


class DocElement:

    def __init__(self, doc_root, line: int):
        self._line = line
        self._doc_root = doc_root

    @property
    def line(self) -> int:
        return self._line

    @line.setter
    def line(self, value: int):
        self._line = value

    @property
    def uri_line(self):
        return f"{self.root.uri}:{self.line}"

    @property
    def root(self):
        return self._doc_root

    def construct(self, data, parent, idx=None, dollar_id=None):
        if dollar_id is None:
            dollar_id = JsonAnchor.empty()

        if isinstance(data, dict):
            if '$ref' in data:
                dref = DocReference(data['$ref'], dollar_id, self.root, data.lc.line)
                return dref
            dobj = DocObject(data, self.root, data.lc.line, dollar_id=dollar_id)
            #for k, v in data.items():
            #    dobj[k] = self.construct(data=v, parent=data, idx=k, dollar_id=dollar_id)
            return dobj
        elif isinstance(data, list):
            da = DocArray(data, self.root, data.lc.line, dollar_id=dollar_id)
            #for i, v in enumerate(data):
            #    da.append(self.construct(data=v, parent=data, idx=i, dollar_id=dollar_id))
            return da
        else:
            if idx is not None:
                if isinstance(parent, dict):
                    dv = DocValue.factory(data, self.root, parent.lc.value(idx)[0])
                    if dv is not None and not isinstance(dv, bool):
                        dv.set_key(idx, parent.lc.key(idx)[0])
                    return dv
                elif isinstance(parent, list):
                    dv = DocValue.factory(data, self.root, parent.lc.item(idx)[0])
                    if dv is not None and not isinstance(dv, bool):
                        dv.set_key(idx, parent.lc.item(idx)[0])
                    return dv
            else:
                return DocValue(data, self.root, line=None)


class DocContainer(DocElement):

    def __init__(self, doc_root, line: int, dollar_id=None):
        if dollar_id is None:
            dollar_id = JsonAnchor.empty()
        self._dollar_id = dollar_id
        super().__init__(doc_root, line)


class DocObject(DocContainer, dict):
    
    def __init__(self, data: dict, doc_root, line: int, dollar_id=None):
        super().__init__(doc_root, line, dollar_id)
        if self.root._dollar_id_token in data:
            self._dollar_id.change_to(data[self.root._dollar_id_token])
            self.root._ref_dictionary.put(self._dollar_id, self)
        for k, v in data.items():
            if k in self.root._parse_options.exclude_dollar_id_parse:
                self.root._dollar_id_token = None
            self[k] = self.construct(data=v, parent=data, idx=k, dollar_id=self._dollar_id.copy())
            if k in self.root._parse_options.exclude_dollar_id_parse:
                self.root._dollar_id_token = self.root._parse_options.dollar_id_token

    def resolve_references(self):
        for k, v in self.items():
            if isinstance(v, DocReference):
                self[k] = v.resolve()
            elif isinstance(v, DocObject):
                v.resolve_references()


class DocArray(DocContainer, list):

    def __init__(self, data: list, doc_root, line: int, dollar_id=None):
        super().__init__(doc_root, line, dollar_id)
        for i, v in enumerate(data):
            self.append(self.construct(data=v, parent=data, idx=i, dollar_id=self._dollar_id.copy()))


class DocReference(DocElement):

    def __init__(self, reference:str, dollar_id:Optional[JsonAnchor], doc_root, line:int):
        super().__init__(doc_root, line)
        self._reference = reference.replace("~0", "~").replace("~1", "/").replace("%25", "%")
        self._dollar_id = dollar_id.copy()
        if self._dollar_id is None:
            self._dollar_id = JsonAnchor.empty()

    @property
    def reference(self):
        return self._reference

    def resolve(self):
        try:
            js_anchor = self._dollar_id.copy().change_to(self._reference)
            node = self.root._ref_dictionary.get(js_anchor)
            return node
        except:
            pass
        reference_parts = self._reference.split('#')
        if len(reference_parts) == 2:
            href, path = reference_parts
        elif len(reference_parts) == 1:
            href = reference_parts[0]
            path = ""
        doc = self.root
        if len(href) > 0:
            doc = self.root.get_doc(href)
        node = doc.get_node(path)
        return node

class DocValue(DocElement):

    def __init__(self, value, doc_root, line: int):
        DocElement.__init__(self, doc_root, line)
        self.data = value
        self.key = None
        self.key_line = None

    @property
    def value(self):
        return self.data

    def set_key(self, key_name, key_line):
        self.key = key_name
        self.key_line = key_line

    def __repr__(self):
        if isinstance(self.data, str):
            return f'"{self.data}"'
        return str(self.data)

    @staticmethod
    def factory(value, doc_root, line: int):
        if isinstance(value, bool):
            return value
        elif isinstance(value, int):
            return DocInteger(value, doc_root, line)
        elif isinstance(value, float):
            return DocFloat(value, doc_root, line)
        elif isinstance(value, str):
            return DocString(value, doc_root, line)
        elif value is None:
            return None
        return DocValue(value, doc_root, line)


class DocInteger(DocValue, int):

    def __new__(cls, value: int, doc_root, line: int):
        di = int.__new__(DocInteger, value)
        di.__init__(value, doc_root, line)
        return di

    def __init__(self, value: int, doc_root, line: int):
        DocValue.__init__(self, value, doc_root, line)


class DocFloat(DocValue, float):

    def __new__(cls, value: float, doc_root, line: int):
        df = float.__new__(DocFloat, value)
        df.__init__(value, doc_root, line)
        return df

    def __init__(self, value: float, doc_root, line: int):
        DocValue.__init__(self, value, doc_root, line)


class DocString(DocValue, str):

    def __new__(cls, value: str, doc_root, line: int):
        # This is stupid and needs to be fixed.
        # It is here to correctly load a poop emoji found
        # in the minLength.json JSON-Schema test data.
        new_value = json.loads(json.dumps(value))
        new_len = len(new_value)
        ds = str.__new__(DocString, new_value)
        ds.__init__(new_value, doc_root, line)
        return ds

    def __init__(self, value: str, doc_root, line: int):
        DocValue.__init__(self, value, doc_root, line)

class Document:
    pass

def create_document(uri, loader: Optional[LoaderBaseClass]=None, options: Optional[ParseOptions]=None):
    if loader is None:
        loader = FilesystemLoader()
    if options is None:
        options = ParseOptions()
    parser = Parser()
    structure = parser.parse_yaml(loader.load(uri))
    base_class = DocObject
    if isinstance(structure, list):
        base_class = DocArray
    elif isinstance(structure, dict) and "$ref" in structure:
        dollar_ref = structure['$ref']
        doc = create_document(dollar_ref, loader, options)
        return doc

    class DocumentRoot(base_class, Document):

        def __init__(self, uri, loader: LoaderBaseClass, options: ParseOptions):
            self._dollar_id_token = options.dollar_id_token
            self._ref_resolution_mode = options.ref_resolution_mode
            self._parse_options = options
            self._uri = uri
            self._loader = loader
            self.parser = Parser()
            self._doc_cache = RemoteDocumentCache(self._loader, self._parse_options)
            structure = self.parser.parse_yaml(loader.load(self._uri))
            self._ref_dictionary = ReferenceDictionary()
            new_dollar_id = JsonAnchor.empty()
            if self._dollar_id_token in structure:
                uri = structure[self._dollar_id_token]
                new_dollar_id = JsonAnchor.from_string(uri)
                self._ref_dictionary.put(new_dollar_id, self)
            super().__init__(structure, self, 0, dollar_id=new_dollar_id)
            if self._ref_resolution_mode == RefResolutionMode.RESOLVE_REFERENCES:
                self.resolve_references()

        @property
        def uri(self):
            return self._uri

        def get_node(self, path):
            path_parts = [ p for p in path.split('/') if len(p) > 0 ]
            node = self
            for part in path_parts:
                try:
                    node = node[part]
                except KeyError:
                    raise PathReferenceResolutionError(self, path)
                except TypeError:
                    raise PathReferenceResolutionError(self, path)
            return node

        def get_doc(self, uri):
            return self._doc_cache.get_doc(uri)

    return DocumentRoot(uri, loader, options)


class RemoteDocumentCache(object):
    _cache = {}
    _loading = set()

    def __init__(self,  loader, options):
        self._loader = loader
        self._parse_options = options

    def get_doc(self, uri):
        if uri in self._loading:
            raise CircularDependencyError(uri)
        if uri not in self._cache:
            self._loading.add(uri)
            doc = create_document(uri, self._loader, self._parse_options)
            self._cache[uri] = doc
            self._loading.remove(uri)
        return self._cache[uri]



from ruamel.yaml import YAML
from ruamel.yaml.scanner import ScannerError

class Parser(object):

    def __init__(self):
        self._yaml = YAML(typ='rt')

    def parse_yaml(self, data: str):
        try:
            structure = self._yaml.load(data)
        except ScannerError as e:
            fixed_data = data.replace("\t", "    ")
            structure = self._yaml.load(fixed_data)
        return structure

    def parse_json(self, data: str):
        return self.parse_yaml(data)



from typing import List

from enum import Enum

class RefResolutionMode(Enum):
    USE_REFERENCES_OBJECTS = 0
    RESOLVE_REFERENCES = 1


class ParseOptions:

    def __init__(self):
        self.ref_resolution_mode:RefResolutionMode = RefResolutionMode.USE_REFERENCES_OBJECTS
        self.dollar_id_token:str = "$id"
        self.exclude_dollar_id_parse:List[str] = ["const", "enum"]



import unittest

import jacobsjsondoc
from jacobsjsondoc.parser import Parser

JSON_WITH_A_TAB = """
{
    "things": [
        { "foo": "bar" },
		{ "hello": "world - this line has a tab character instead of spaces" }
    ]
}
"""

class TestParseJson(unittest.TestCase):

    def setUp(self):
        self.parser = Parser()

    def test_parse_json_with_tab(self):
        doc = self.parser.parse_json(JSON_WITH_A_TAB)
        self.assertIsInstance(doc, dict)
        self.assertIsInstance(doc["things"], list)



from setuptools import setup, find_packages
import jacobsjsondoc._version as _version

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="jacobs-json-doc",
    version=_version.__version__,
    description='A JSON/YAML loader',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Jacob Brunson",
    author_email="pypi@jacobbrunson.com",
    url="https://github.com/pearmaster/jacobs-json-doc",
    license='GPLv2',
    packages=find_packages(),
    install_requires=[
        "ruamel.yaml",
    ],
    keywords='conversion',
    classifiers= [
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.7',
)


import unittest
from .context import jacobsjsondoc
from jacobsjsondoc.reference import JsonAnchor, ReferenceDictionary
from jacobsjsondoc.document import create_document, DocReference
from jacobsjsondoc.loader import PrepopulatedLoader
from jacobsjsondoc.options import ParseOptions, RefResolutionMode
import json

SAMPLE_DOCUMENT = {
    "$id": "http://example.com/schema.json",
    "type": "object",
    "properties": {
        "foo": {
            "$ref": "#fooprop",
        },
        "bar": {
            "$id": "#barprop",
            "type": "integer",
        }
    },
    "objects": {
        "fooProperty": {
            "$id": "#fooprop",
            "type": "string",
        }
    }
}

class TestJsonReferenceObject(unittest.TestCase):

    def test_reference_from_uri(self):
        uri = "http://example.com/schema.json#/definition/food"
        ref = JsonAnchor.from_string(uri)
        self.assertEquals(ref.uri, "http://example.com/schema.json")

    def test_references_equal(self):
        uri = "http://example.com/schema.json#/definition/food"
        ref1 = JsonAnchor.from_string(uri)
        ref2 = JsonAnchor.from_string(uri)
        self.assertEquals(ref1, ref2)
        ref3 = ref1.copy()
        self.assertEquals(ref2, ref3)

    def test_reference_buildup(self):
        base_uri = "http://example.com/myschema.json"
        ref = JsonAnchor.from_string(base_uri)
        change_path_id = "/other/schema.json"
        ref.change_to(JsonAnchor.from_string(change_path_id))
        self.assertEquals(ref.uri, "http://example.com/other/schema.json")
        add_fragment_id = "#func"
        ref.change_to(JsonAnchor.from_string(add_fragment_id))
        ref_repr = repr(ref)
        self.assertEquals(ref_repr, "http://example.com/other/schema.json#func")
        ref2 = JsonAnchor.from_string(ref_repr)
        self.assertEquals(ref, ref2)

class TestReferenceDictionary(unittest.TestCase):

    def setUp(self):
        self.data1 = {
            "A": {
                "B": 1,
                "C": [2,3,4,5]
            },
            "D": False
        }

    def test_reference_lookup(self):
        source_uri = "example"
        rd = ReferenceDictionary()
        rd.put(source_uri, self.data1)
        ref = JsonAnchor.from_string(source_uri)
        node_out = rd[ref]
        self.assertEqual(self.data1, node_out)
        ref.change_to(JsonAnchor.from_string("#A/B"))
        rd[ref] = self.data1['A']['B']
        fragment_uri = "example#A/B"
        self.assertEqual(rd.get(fragment_uri), 1)


class TestIdTagging(unittest.TestCase):

    def setUp(self):
        self.data = SAMPLE_DOCUMENT
        ppl = PrepopulatedLoader()
        ppl.prepopulate(self.data["$id"], json.dumps(self.data))
        self.doc = create_document(uri=self.data["$id"], loader=ppl)
    
    def test_root_has_correct_id(self):
        self.assertEquals(self.doc._dollar_id.uri, self.data["$id"])

    def test_bar_has_correct_id(self):
        self.assertEquals(self.doc['properties']['bar']._dollar_id, "http://example.com/schema.json#barprop")

    def test_fooproperty_has_correct_id(self):
        self.assertEquals(self.doc['objects']['fooProperty']._dollar_id, "http://example.com/schema.json#fooprop")

    def test_dictionary_contents(self):
        print(self.doc._ref_dictionary)
        self.assertEqual(len(self.doc._ref_dictionary), 3)

    def test_dictionary_has_barprop(self):
        barprop = self.doc._ref_dictionary.get("http://example.com/schema.json#barprop")
        self.assertEquals(barprop['$id'], "#barprop")
        self.assertEquals(barprop['type'], "integer")
    
DOUBLE_REFERENCE_DOC = """
{
    "definitions": {
        "item": {
            "type": "array",
            "additionalItems": false,
            "items": [
                { "$ref": "#/definitions/sub-item" },
                { "$ref": "#/definitions/sub-item" }
            ]
        },
        "sub-item": {
            "type": "object",
            "required": ["foo"]
        }
    },
    "type": "array",
    "additionalItems": false,
    "items": [
        { "$ref": "#/definitions/item" },
        { "$ref": "#/definitions/item" },
        { "$ref": "#/definitions/item" }
    ]
}
"""

class TestDoubleRef(unittest.TestCase):

    def setUp(self):
        self.data = DOUBLE_REFERENCE_DOC
        ppl = PrepopulatedLoader()
        ppl.prepopulate(1, self.data)
        self.doc = create_document(uri=1, loader=ppl)

    def test_is_a_reference(self):
        self.assertIsInstance(self.doc['items'][0], DocReference)
        resolved = self.doc['items'][0].resolve()
        self.assertEqual(resolved['type'], "array")
        self.assertIsInstance(resolved['items'], list)

class TestIdTrouble(unittest.TestCase):

    def setUp(self):
        data_text = """
        "schema": {
            "definitions": {
                "id_in_enum": {
                    "enum": [
                        {
                          "id": "https://localhost:1234/my_identifier.json",
                          "type": "null"
                        }
                    ]
                },
                "real_id_in_schema": {
                    "id": "https://localhost:1234/my_identifier.json",
                    "type": "string"
                },
                "zzz_id_in_const": {
                    "const": {
                        "id": "https://localhost:1234/my_identifier.json",
                        "type": "null"
                    }
                }
            },
            "anyOf": [
                { "$ref": "#/schema/definitions/id_in_enum" },
                { "$ref": "https://localhost:1234/my_identifier.json" }
            ]
        }
        """
        ppl = PrepopulatedLoader()
        ppl.prepopulate(1, data_text)
        options = ParseOptions()
        options.ref_resolution_mode = RefResolutionMode.USE_REFERENCES_OBJECTS
        options.dollar_id_token = "id"
        self.doc = create_document(uri=1, loader=ppl, options=options)

    def test_ref_points_to_correct_id(self):
        first_anyof_ref = self.doc["schema"]["anyOf"][0]
        self.assertIsInstance(first_anyof_ref, DocReference)
        second_anyof_ref = self.doc["schema"]["anyOf"][1]
        self.assertIsInstance(second_anyof_ref, DocReference)

        first_resolved = first_anyof_ref.resolve()
        second_resolved = second_anyof_ref.resolve()
        


from abc import ABC, abstractmethod

class LoaderBaseClass(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def load(self, uri) -> str:
        pass


class FilesystemLoader(LoaderBaseClass):

    def __init__(self):
        super().__init__()

    def load(self, uri: str) -> str:
        return open(uri).read()


class PrepopulatedLoader(LoaderBaseClass):

    def __init__(self):
        super().__init__()
        self._documents = {}

    def prepopulate(self, uri, source):
        self._documents[uri] = source

    def load(self, uri: str) -> str:
        return self._documents[uri]


import unittest
from collections import UserDict
from jacobsjsondoc.loader import PrepopulatedLoader
from jacobsjsondoc.document import create_document, RefResolutionMode, DocReference, DocValue, DocObject, CircularDependencyError
from jacobsjsondoc.options import ParseOptions, RefResolutionMode

SIMPLE_YAML = """
jacob:
    brunson:
        - 1
        - 2
        - true
        - isa:
            nice guy
    food:
        true
def:
    foo: "this is a foo string"
"""

SIMPLE_WITH_INTEGER = """
thevalue: 42
"""

YAML_WITH_REF = """
house:
    var: "this is the value of the var"
    local:
        $ref: "#/house/var"
    remote:
        $ref: "remote#/def/foo"
"""

ANOTHER_YAML_WITH_REF = """
colorado:
    denver: 1
    springs: 
        $ref: "middle#/house"
"""

YAML_TYPES = """
myobject:
    myint: 10
    myfloat: 3.14159
    mystring: string
    mytrue: true
    myfalse: false
    mynull: null
"""


class TestDocument(unittest.TestCase):

    def test_lines(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate(None, SIMPLE_YAML)
        doc = create_document(uri=None, loader=ppl)
        
        self.assertEqual(doc['jacob'].line, 2)
        self.assertEqual(doc['jacob']['brunson'].line, 3)
        self.assertEquals(doc['jacob']['brunson'][0].value, 1)
        self.assertEqual(doc['jacob']['brunson'][0].line, 3)

    def test_local_ref_use_reference_objects(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate("yaml_with_ref", YAML_WITH_REF)
        options = ParseOptions()
        options.ref_resolution_mode = RefResolutionMode.USE_REFERENCES_OBJECTS
        doc = create_document(uri="yaml_with_ref", loader=ppl, options=options)
        self.assertIsInstance(doc['house']['local'], DocReference)
        node = doc['house']['local'].resolve()
        self.assertIsInstance(node, DocValue)
        self.assertEqual(node.value, "this is the value of the var")

    def test_local_ref_resolve_references(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate("local", YAML_WITH_REF)
        ppl.prepopulate("remote", SIMPLE_YAML)
        options = ParseOptions()
        options.ref_resolution_mode = RefResolutionMode.RESOLVE_REFERENCES
        doc = create_document(uri="local", loader=ppl, options=options)
        self.assertIsInstance(doc['house']['local'], DocValue)
        self.assertEqual(doc['house']['local'].value, "this is the value of the var")

    def test_remote_ref_use_reference_objects(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate("local", YAML_WITH_REF)
        ppl.prepopulate("remote", SIMPLE_YAML)
        options = ParseOptions()
        options.ref_resolution_mode = RefResolutionMode.USE_REFERENCES_OBJECTS
        doc = create_document(uri="local", loader=ppl, options=options)
        self.assertIsInstance(doc['house']['remote'], DocReference)
        node = doc['house']['remote'].resolve()
        self.assertIsInstance(node, DocValue)
        self.assertEqual(node.value, "this is a foo string")
        self.assertEqual(node.line, 11)

    def test_remote_ref_resolve_references(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate("local", YAML_WITH_REF)
        ppl.prepopulate("remote", SIMPLE_YAML)
        options = ParseOptions()
        options.ref_resolution_mode = RefResolutionMode.RESOLVE_REFERENCES
        doc = create_document(uri="local", loader=ppl, options=options)
        self.assertIsInstance(doc['house']['remote'], DocValue)
        self.assertEqual(doc['house']['remote'].value, "this is a foo string")

    def test_3_layer_resolve_references(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate("middle", YAML_WITH_REF)
        ppl.prepopulate("remote", SIMPLE_YAML)
        ppl.prepopulate("local", ANOTHER_YAML_WITH_REF)
        options = ParseOptions()
        options.ref_resolution_mode = RefResolutionMode.RESOLVE_REFERENCES
        doc = create_document(uri="local", loader=ppl, options=options)
        self.assertIsInstance(doc['colorado']['springs'], DocObject)
        self.assertIsInstance(doc['colorado']['springs']['var'], DocValue)
        self.assertEqual(doc['colorado']['springs']['var'].value, "this is the value of the var")
        self.assertEqual(doc['colorado']['springs']['var'].uri_line, "middle:2")
        self.assertIsInstance(doc['colorado']['springs']['remote'], DocValue)
        self.assertEqual(doc['colorado']['springs']['remote'].value, "this is a foo string")
        self.assertEqual(doc['colorado']['springs']['remote'].uri_line, "remote:11")

    def test_circular_dependency(self):
        yaml1 = """
        outer:
            inner:
                $ref: "two#/foo"
        """
        yaml2 = """
        foo:
            bar:
                $ref: "one#/outer"
        """
        ppl = PrepopulatedLoader()
        ppl.prepopulate("one", yaml1)
        ppl.prepopulate("two", yaml2)
        options = ParseOptions()
        options.ref_resolution_mode = RefResolutionMode.RESOLVE_REFERENCES
        with self.assertRaises(CircularDependencyError):
            doc = create_document(uri="one", loader=ppl, options=options)


class TestDocumentTypes(unittest.TestCase):

    def setUp(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate(None, YAML_TYPES)
        self.doc = create_document(uri=None, loader=ppl)

    def test_userdict(self):
        self.assertIsInstance(self.doc['myobject'], dict)

    def test_integer(self):
        self.assertEqual(self.doc['myobject']['myint'], 10)
        self.assertIsInstance(self.doc['myobject']['myint'], int)
        self.assertEqual(self.doc['myobject']['myint'].line, 2)

    def test_float(self):
        self.assertEqual(self.doc['myobject']['myfloat'], 3.14159)
        self.assertIsInstance(self.doc['myobject']['myfloat'], float)
        self.assertEqual(self.doc['myobject']['myfloat'].line, 3)

    def test_string(self):
        self.assertEqual(self.doc['myobject']['mystring'], "string")
        self.assertIsInstance(self.doc['myobject']['mystring'], str)
        self.assertEqual(self.doc['myobject']['mystring'].line, 4)

    def test_boolean_true(self):
        self.assertEqual(self.doc['myobject']['mytrue'], True)

    def test_boolean_false(self):
        self.assertEqual(self.doc['myobject']['myfalse'], False)

    def test_null(self):
        self.assertEqual(self.doc['myobject']['mynull'], None)
        self.assertIsInstance(self.doc['myobject']['mynull'], type(None))

class TestDocumentSimpleTypes(unittest.TestCase):

    def test_integer(self):
        ppl = PrepopulatedLoader()
        ppl.prepopulate(None, SIMPLE_WITH_INTEGER)
        self.doc = create_document(uri=None, loader=ppl)
        self.assertEqual(self.doc['thevalue'], 42)
        self.assertIsInstance(self.doc['thevalue'], int)
        self.assertEqual(self.doc['thevalue'].line, 1)

class TestDocumentReference(unittest.TestCase):

    def setUp(self):
        doc_text = """
        {
            "$ref": "http://example.org/schema"
        }
        """
        remote = """
        { "type": "integer" }
        """
        ppl = PrepopulatedLoader()
        ppl.prepopulate("local", doc_text)
        ppl.prepopulate("http://example.org/schema", remote)
        self.doc = create_document(uri="local", loader=ppl)

    def test_loaded_reference(self):
        self.assertIsInstance(self.doc, dict)

    def test_correct_remote(self):
        self.assertTrue("type" in self.doc)
        self.assertEqual(self.doc["type"], "integer")

if __name__ == '__main__':
    unittest.main()


import unittest

import jacobsjsondoc
from jacobsjsondoc.document import DocObject

SUPER_SIMPLE_JSON = """
{
    "Hello": "Grand World"
}
"""

class TestQuickParse(unittest.TestCase):

    def test_quick_parse_json(self):
        doc = jacobsjsondoc.parse(SUPER_SIMPLE_JSON)
        self.assertIsInstance(doc, DocObject)